#include <stdio.h>
#include <mpi.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <memory.h>

#define TAG 1
#define SUBTAG 2

int main(int argc, char ** argv) {

//The only argument here is the range, R, of the random numbers.
//Each node will generate a random number, and the whole list of

	int my_rank, world_size;
	int my_sub_rank, sub_group_size;

        char message[256], name[80];

        MPI_Status status; // for checking info on messages sent and recieved
	MPI_Group MPI_GROUP_WORLD, sub_group; // MPI_GROUP_WORLD will be large group, subgroup will be all but node 0
	MPI_Comm sub_comm; // create a new communication variable for the sub_group

	int j = 0;
	int k = 0;
	int R = atoi(argv[1]);
	int guess_counter = 0;

	int sub_recv = -1; // for setting up send/recv among subgroup
	int sub_send = -1; // for setting up send/recv among subgroup
	int sub_randval = -1; // ditto

	int sub_test = 0; // for testing MPI_Bcast in the subgroup

	MPI_Init(&argc, &argv); /* note that argc and argv are passed by address */

        MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);
        MPI_Comm_size(MPI_COMM_WORLD, &world_size);

	int *ranks = NULL;

	// we are creating a ranks array that will determine who gets put in the sub_group
	// if your rank appears in this array, you go to the subgroup
	ranks = (int *)malloc((world_size-1)*sizeof(int));

	for(j=1; j<world_size; j++)
	{
		ranks[j-1]=j;
	}	

	// If we put the random number generator stuff in here, maybe it can be more "random"
	int seed;
	double rand1 = 0.0;
	double temp = 0.0;
	seed = time(NULL) + my_rank;
	srand(seed);
	int randval;
	int sentrandval;

	int gono = 1;
	
	// Since outside of an if (my_rank == ...), all nodes do this number generation:
	rand1 = rand();
	temp = rand1 / (double) RAND_MAX;
	randval = R*temp;
	printf("Node %d: My random number is %d \n", my_rank, randval);

	// Same here, all nodes read MPI_Gather, but the result is sent to node 0.
	// MPI_Gather(&randval, 1, MPI_INT, RandomArray, 1, MPI_INT, 0, MPI_COMM_WORLD);

	// After MPI_Init, we now do an MPI_Comm_Group so that we know what the original group is:
	MPI_Comm_group(MPI_COMM_WORLD, &MPI_GROUP_WORLD);

	// Now using that group, we create the subgroup. Note we need MPI_COMM_WORLD, new member amount, the ranks, and new group name
	MPI_Group_incl(MPI_GROUP_WORLD, world_size-1, ranks, &sub_group);
	
	MPI_Comm_create(MPI_COMM_WORLD, sub_group, &sub_comm);

	if (sub_comm != MPI_COMM_NULL)
	{
        	MPI_Comm_rank(sub_comm, &my_sub_rank); // generate new ranks in the subgroup only
        	MPI_Comm_size(sub_comm, &sub_group_size); // generate new world size for the subgroup only
		printf("Node %d: I am in the subgroup, with new rank %d in a subgroup size of %d \n", my_rank, my_sub_rank, sub_group_size);

		sub_recv = (my_sub_rank+sub_group_size-1)%sub_group_size;
		sub_send = (my_sub_rank+sub_group_size+1)%sub_group_size;
		printf("Node %d, subNode %d: sub_recv is %d, sub_send is %d \n", my_rank, my_sub_rank, sub_recv, sub_send);

		printf("Node %d, subNode %d: Sending subNode %d in the subgroup my random value of %d\n", my_rank, my_sub_rank, sub_send, randval);
		MPI_Send(&randval, 1, MPI_INT, sub_send, SUBTAG, sub_comm);

		MPI_Recv(&sub_randval, 1, MPI_INT, sub_recv, SUBTAG, sub_comm, &status);
		printf("Node %d, subNode %d: Receiving from subNode %d in the subgroup the random value of %d\n", my_rank, my_sub_rank, sub_recv, sub_randval);

		if (my_sub_rank == 0)
		{
			printf("Node %d, subnode %d: Changing sub_randval to new value of 11\n", my_rank, my_sub_rank);
			sub_randval = 11;
		}

		printf("Node %d, subnode %d: Before Bcast, sub_randval  = %d \n", my_rank, my_sub_rank, sub_randval);
		MPI_Bcast(&sub_randval, 1, MPI_INT, 0, sub_comm);
		printf("Node %d, subnode %d: After Bcast, sub_randval  = %d \n", my_rank, my_sub_rank, sub_randval);
	}

    MPI_Finalize();

}
